library(rSPDE)
library(INLA)
library(splancs)
library(viridis)
library(excursions)
library(RColorBrewer)

#############################################################
# Define some location functions to simplify the code later #
#############################################################

# Function that computes the CRPS for a Gaussian distribution
CRPS <- function(y, mu, sigma) {
  mu_y <- as.matrix(mu - y)
  return(sigma * sqrt(2 / pi) * exp(-mu_y^2 / (2 * sigma^2))
    + mu_y * (1 - 2 * pnorm(-mu_y / sigma)) - sigma / sqrt(pi))
}

# Function that computes the log-score for a Gaussian distribution
LS <- function(y, mu, sigma) {
  return(dnorm(y, mean = mu, sd = sigma, log = TRUE))
}

# Cross-validation functino for evaluation of the predictive performance
crossval.loop <- function(op, ind, Y, A, sigma_est) {
  n.rep <- dim(Y)[2]
  mse_i <- crps_i <- brps_i <- ls_i <- matrix(0, 10, n.rep)
  for (i in 1:10) {
    if (length(sigma_est) == 1) {
      se <- sigma_est
      se2 <- sigma_est
    } else {
      se <- sigma_est[ind[, i] == 1, ]
      se2 <- sigma_est[ind[, i] == 0, ]
    }
    pred <- predict(op,
      A = A[ind[, i] == 1, ],
      Aprd = A[ind[, i] == 0, ],
      Y = Y[ind[, i] == 1, ],
      sigma.e = se,
      compute.variances = TRUE
    )

    for (j in 1:n.rep) {
      mse_i[i, j] <- mean((Y[ind[, i] == 0, j] - pred$mean[, j])^2)
      crps_i[i, j] <- mean(CRPS(
        Y[ind[, i] == 0, j], pred$mean[, j],
        sqrt(pred$variance + se2^2)
      ))
      ls_i[i, j] <- mean(LS(
        Y[ind[, i] == 0, j], pred$mean[, j],
        sqrt(pred$variance + se2^2)
      ))
    }
  }
  return(list(
    rmse = sqrt(mean(c(mse_i))),
    crps = mean(c(crps_i)),
    brps = mean(c(brps_i)),
    ls = mean(c(ls_i))
  ))
}

# Function that assembles a matrix with basis functions for the non-stationary parameters
basis.matrix <- function(x, y, xd, yd, n) {
  xs <- (x - min(xd)) / (max(xd) - min(xd))
  ys <- (y - min(yd)) / (max(yd) - min(yd))
  M <- matrix(1, length(x), 1)
  for (i in 1:n) {
    for (j in 1:n) {
      M <- cbind(M, sin(i * pi * xs) * sin(j * pi * ys))
      M <- cbind(M, cos(i * pi * xs) * cos(j * pi * ys))
      M <- cbind(M, sin(i * pi * xs) * cos(j * pi * ys))
      M <- cbind(M, cos(i * pi * xs) * sin(j * pi * ys))
    }
  }
  return(M)
}

################################
# Load data and construct mesh #
################################
load("precip.data.Rdata")
set.seed(1)

mesh <- inla.mesh.2d(
  loc = loc,
  max.edge = c(2, 3),
  offset = c(-0.05, -0.1),
  min.angle = 21
)

fem <- inla.fmesher.smorg(mesh$loc,
  mesh$graph$tv,
  fem = 2,
  output = list("c0", "c1", "g1")
)

A <- inla.spde.make.A(mesh, loc = loc)
C <- fem$c0
G <- fem$g1

n_rep <- dim(Y)[2]
N <- dim(loc)[1]

# create cross validation indices
ind <- matrix(0, N, 10)
for (i in 1:10) {
  ind_v <- sample(N, replace = FALSE)
  ind_v <- ind_v[1:round(N * 0.1)]
  ind[ind_v, i] <- 1
}
n <- mesh$n

# construct altitude covariate for field by averaging the altitude over each triangle in the mesh
A.alt <- inla.spde.make.A(mesh, loc = loc.alt)
alt.mesh <- as.vector(t(A.alt) %*% c(alt)) / colSums(A.alt)
alt.mesh[is.na(alt.mesh)] <- 0

###################################
# stationary non-fractional model #
###################################

mlik <- function(theta, Y, G, C, A) {
  kappa = exp(theta[1])
  tau = exp(theta[2])
  nu = 1
  sigma.e = exp(theta[3])
  return(-spde.matern.loglike(kappa, tau, nu, sigma.e, Y = Y, 
                              G = G, C = C, A = A, d = 2, m = 1))
}

cat("Estimate stationary non-fractional model\n")
vars <- apply(Y, 1, var)
theta0 <- log(c(sqrt(8) / 10, mean(sqrt(vars)), 0.001))
pars <- optim(theta0, mlik,
  Y = Y, G = G, C = C, A = A, method = "L-BFGS-B",
  control = list(REPORT = 1, trace = 1)
)
like_s1 <- pars$value
theta_s1 <- pars$par

kappa_est <- exp(theta_s1[1])
tau_est <- exp(theta_s1[2])
sigma_est <- exp(theta_s1[3])
rho_est <- sqrt(8) / kappa_est
phi_est <- sqrt(gamma(1) / (tau_est^2 * gamma(2) * (4 * pi) * kappa_est^(2)))

cat("Cross-validation for stationary non-fractional model\n")
op <- spde.matern.operators(kappa = kappa_est, tau = tau_est, nu = 1, G = G, C = C, d = 2, m = 1)
res.s1 <- crossval.loop(op, ind, Y, A, sigma_est)

cat("Likelihood timing for stationary non-fractional model\n")
tmp <- rep(0, 100)
ptm <- proc.time()[3]
for (i in 1:100) tmp[i] <- mlik(theta = theta_s1, Y = Y, G = G, C = C, A = A)
t.s1 <- (proc.time()[3] - ptm) / 100

###############################
# stationary fractional model #
###############################

mlik <- function(theta, Y, G, C, A) {
  kappa = exp(theta[1])
  tau = exp(theta[2])
  nu = exp(theta[4])
  sigma.e = exp(theta[3])
  return(-spde.matern.loglike(kappa, tau, nu, sigma.e, Y = Y, G = G, C = C, A = A, d = 2, m = 1))
}
cat("Estimate stationary fractional model\n")
theta0 <- c(theta_s1, log(3 / 4))
pars <- optim(theta0, mlik,
  Y = Y, G = G, C = C, A = A, method = "L-BFGS-B",
  control = list(REPORT = 1, trace = 1)
)

like_sr <- pars$value
theta_sr <- pars$par

kappa_est <- exp(theta_sr[1])
tau_est <- exp(theta_sr[2])
sigma_est <- exp(theta_sr[3])
nu_est <- exp(theta_sr[4])
rho_est <- sqrt(8 * nu_est) / kappa_est
phi_est <- sqrt(gamma(nu_est) / (tau_est^2 * gamma(nu_est + 1) * (4 * pi) * kappa_est^(2 * nu_est)))

cat("Cross-validation for stationary fractional model\n")
op <- spde.matern.operators(kappa = kappa_est, tau = tau_est, nu = nu_est, G = G, C = C, d = 2, m = 1)
res.sr <- crossval.loop(op, ind, Y, A, sigma_est)

cat("Likelihood timing for stationary fractional model\n")
tmp <- rep(0, 100)
ptm <- proc.time()[3]
for (i in 1:100) tmp[i] <- mlik(theta = theta_sr, Y = Y, G = G, C = C, A = A)
t.sr <- (proc.time()[3] - ptm) / 100

#######################################
# non-stationary non-fractional model #
#######################################
    
mlik <- function(theta, Y, G, C, A, B) {
  n_basis <- dim(B)[2]
  kp <- exp(B %*% theta[1:n_basis])
  tau <- exp(B %*% theta[(n_basis + 1):(2 * n_basis)])
  sigma <- exp(theta[2 * n_basis + 1])
  return(-spde.matern.loglike(kp, tau, 1, sigma, Y = Y, G = G, C = C, A = A, d = 2, m = 1))
}

cat("Estimate non-stationary non-fractional model\n")
B <- basis.matrix(mesh$loc[, 1], mesh$loc[, 2], loc[, 1], loc[, 2], 2)
B <- cbind(B, alt.mesh / 1000) # add altitude covariate in km scale
n_basis <- dim(B)[2]

theta0 <- c(theta_s1[1], rep(0, n_basis - 1), theta_s1[2], rep(0, n_basis - 1), theta_s1[3])
pars <- optim(theta0, mlik,
  Y = Y, G = G, C = C, A = A, B = B, method = "L-BFGS-B",
  control = list(REPORT = 1, trace = 1, maxit = 500)
)
pars <- optim(pars$par, mlik,
  Y = Y, G = G, C = C, A = A, B = B, method = "L-BFGS-B",
  control = list(REPORT = 1, trace = 1, maxit = 500)
)

like_n1 <- pars$value
theta_n1 <- pars$par

kappa_est <- exp(B %*% theta_n1[1:n_basis])
tau_est <- exp(B %*% theta_n1[(n_basis + 1):(2 * n_basis)])
sigma_est <- exp(theta_n1[2 * n_basis + 1])

cat("Cross-validation for non-stationary non-fractional model\n")
op <- spde.matern.operators(kappa_est, tau_est, 1, G, C, d = 2, m = 1)
res.n1 <- crossval.loop(op, ind, Y, A, sigma_est)

cat("Likelihood timing for non-stationary non-fractional model\n")
tmp <- rep(0, 100)
ptm <- proc.time()[3]
for (i in 1:100) tmp[i] <- mlik(theta = theta_n1, Y = Y, G = G, C = C, A = A, B = B)
t.n1 <- (proc.time()[3] - ptm) / 100

###################################
# non-stationary fractional model #
###################################

mlik <- function(theta, Y, G, C, A, B) {
  n_basis <- dim(B)[2]
  kp <- exp(B %*% theta[1:n_basis])
  tau <- exp(B %*% theta[(n_basis + 1):(2 * n_basis)])
  sigma <- exp(theta[2 * n_basis + 1])
  nu <- exp(theta[2 * n_basis + 2])
  return(-spde.matern.loglike(kp, tau, nu, sigma, Y = Y, G = G, C = C, A = A))
}

cat("Estimate non-stationary fractional model\n")

theta0 <- c(theta_sr[1], rep(0, n_basis - 1), theta_sr[2], rep(0, n_basis - 1), theta_sr[3:4])
pars <- optim(theta0, mlik,
  Y = Y, G = G, C = C, A = A, B = B, method = "L-BFGS-B",
  control = list(REPORT = 1, trace = 1, maxit = 500)
)
pars <- optim(pars$par, mlik,
  Y = Y, G = G, C = C, A = A, B = B, method = "L-BFGS-B",
  control = list(REPORT = 1, trace = 1, maxit = 500)
)

like_nr <- pars$value
theta_nr <- pars$par

kappa_est <- exp(B %*% theta_nr[1:n_basis])
tau_est <- exp(B %*% theta_nr[(n_basis + 1):(2 * n_basis)])
sigma_est <- exp(theta_nr[2 * n_basis + 1])
nu_est <- exp(theta_nr[2 * n_basis + 2])

cat("Cross-validation for non-stationary fractional model\n")
op <- spde.matern.operators(kappa_est, tau_est, nu_est, G, C, d = 2, m = 1)
res.nr <- crossval.loop(op, ind, Y, A, sigma_est)

cat("Likelihood timing for non-stationary fractional model\n")
tmp <- rep(0, 100)
ptm <- proc.time()[3]
for (i in 1:100) tmp[i] <- mlik(theta = theta_nr, Y = Y, G = G, C = C, A = A, B = B)
t.nr <- (proc.time()[3] - ptm) / 100

####################################################
# Display results
####################################################

results <- data.frame(
  like = -c(like_s1, like_sr, like_n1, like_nr),
  rmse = 100 * c(res.s1$rmse, res.sr$rmse, res.n1$rmse, res.nr$rmse),
  crps = 100 * c(res.s1$crps, res.sr$crps, res.n1$crps, res.nr$crps),
  ls = 100 * c(res.s1$ls, res.sr$ls, res.n1$ls, res.nr$ls),
  time = c(t.s1, t.sr, t.n1, t.nr),
  row.names = c(
    "stationary v=1", "stationary fractional",
    "non-stationary v=1", "non-stationary fractional"
  )
)
print(results, digits = 4)


################
# Plot results #
################
tmp <- matrix(NA, 147, 116)
tmp[xy.in] <- Y[, 1]
cmap <- colorRampPalette(brewer.pal(9, "YlGnBu"))(100) [1:80]

# plot data and mesh
plot(mesh,
  draw.segments = FALSE, main = "", edge.color = rgb(0.4, 0.4, 0.4),
  xlim = c(-134, -45)
)
image.plot(matrix(loc.g[, 1], 147, 116), matrix(loc.g[, 2], 147, 116), tmp,
  col = cmap, add = TRUE,
  transparent.color = "transparent", legend.mar = 3.5
)
lines(border.points[, 1], border.points[, 2], lwd = 2, type = "l", col = "black")
axis(1, seq(-140, -60, by = 10), labels = seq(-140, -60, by = 10))
axis(2, seq(10, 60, by = 10), labels = seq(10, 60, by = 10))
title(xlab = "Longitude", ylab = "Latitude")

# plot covariates
proj <- inla.mesh.projector(mesh, dims = c(200, 300))
cmap <- colorRampPalette(brewer.pal(9, "Reds"))(100)

xy.in <- inout(proj$lattice$loc, border.points)
nxy <- c(length(proj$x), length(proj$y))
B.p <- inla.mesh.project(proj, B[, 18])
B.p[!xy.in] <- NA
image(proj$x, proj$y, B.p,
  col = cmap,
  xlim = c(-127, -65), ylim = c(25, 50), xlab = "",
  ylab = "", ann = FALSE, xaxt = "n", yaxt = "n"
)
lines(border.points[, 1], border.points[, 2], lwd = 2, type = "l", col = "black")

cmap <- viridis(100)
for (i in 1:8) {
  B.p <- inla.mesh.project(proj, B[, i + 1])
  B.p[!xy.in] <- NA
  image(proj$x, proj$y, B.p,
    col = cmap,
    xlim = c(-127, -65), ylim = c(25, 50), zlim = c(-1, 1), xlab = "",
    ylab = "", ann = FALSE, xaxt = "n", yaxt = "n"
  )
  lines(border.points[, 1], border.points[, 2], lwd = 2, type = "l", col = "black")
}


# plot standard deviation and covariances
kappa_est <- exp(B %*% theta_nr[1:n_basis])
tau_est <- exp(B %*% theta_nr[(n_basis + 1):(2 * n_basis)])
sigma_est <- exp(theta_nr[2 * n_basis + 1])
nu_est <- exp(theta_nr[2 * n_basis + 2])

op <- spde.matern.operators(kappa_est, tau_est, nu_est, G, C, d = 2, m = 1)
pred.var <- diag(op$Pr %*% inla.qsolve(op$Q, t(op$Pr)))

lon.grid <- seq(from = min(border.points[, 1]) - 1, to = max(border.points[, 1]), by = 4)
lat.grid <- seq(from = min(border.points[, 2]) - 0.5, to = max(border.points[, 2]) - 2, length.out = 7)
m.grid <- inla.mesh.lattice(x = lon.grid, y = lat.grid)
grid.loc <- m.grid$loc
xy.in <- inout(grid.loc, border.points)
g.loc <- grid.loc[xy.in, ]
A.g <- inla.spde.make.A(mesh, loc = g.loc)


plot(border.points[, 1], border.points[, 2], lwd = 2, type = "l", col = "black", xlab = "", ylab = "", axes = FALSE, asp = 1)
box()
points(g.loc[, 1], g.loc[, 2], col = 2, pch = 4, cex = 0.5)

for (i in 1:dim(g.loc)[1]) {
  cov1 <- Sigma.mult(op,t(A.g[i, , drop = FALSE]))
  con1 <- tricontourmap(mesh, z = as.vector(cov1), levels = as.vector(0.7 * max(cov1)))
  plot(con1$contour, add = TRUE, col = "blue")
}

proj <- inla.mesh.projector(mesh, dims = c(200, 300))
cmap <- rev(colorRampPalette(brewer.pal(9, "YlGnBu"))(100))
xy.in <- inout(proj$lattice$loc, border.points)
nxy <- c(length(proj$x), length(proj$y))
std.p.nf <- inla.mesh.project(proj, sqrt(c(pred.var)))
std.p.nf[!xy.in] <- NA

kappa_est <- exp(B %*% theta_n1[1:n_basis])
tau_est <- exp(B %*% theta_n1[(n_basis + 1):(2 * n_basis)])
sigma_est <- exp(theta_n1[2 * n_basis + 1])
nu_est <- 1

op <- spde.matern.operators(kappa_est, tau_est, nu_est, G, C, d = 2, m = 1)
pred.var <- diag(op$Pr %*% inla.qsolve(op$Q, t(op$Pr)))

plot(border.points[, 1], border.points[, 2], lwd = 2, type = "l", col = "black", xlab = "", ylab = "", axes = FALSE, asp = 1)
box()
points(g.loc[, 1], g.loc[, 2], col = 2, pch = 4, cex = 0.5)

for (i in 1:dim(g.loc)[1]) {
  cov1 <- Sigma.mult(op,t(A.g[i, , drop = FALSE]))
  con1 <- tricontourmap(mesh, z = as.vector(cov1), levels = as.vector(0.7 * max(cov1)))
  plot(con1$contour, add = TRUE, col = "blue")
}

std.p.n1 <- inla.mesh.project(proj, sqrt(c(pred.var)))
std.p.n1[!xy.in] <- NA

p.range.nf <- c(min(std.p.nf[!is.na(std.p.nf)]), max(std.p.nf[!is.na(std.p.nf)]))
p.range.n1 <- c(min(std.p.n1[!is.na(std.p.n1)]), max(std.p.n1[!is.na(std.p.n1)]))

p.var.range <- c(min(c(p.range.n1[1], p.range.nf[1])), max(c(p.range.n1[2], p.range.nf[2])))

image(proj$x, proj$y, std.p.nf,
  col = cmap, xlab = "",
  ylab = "", ann = FALSE, xaxt = "n", yaxt = "n", zlim = p.var.range
)
lines(border.points[, 1], border.points[, 2], lwd = 2, type = "l", col = "black")

image(proj$x, proj$y, std.p.n1,
  col = cmap, xlab = "",
  ylab = "", ann = FALSE, xaxt = "n", yaxt = "n", zlim = p.var.range
)
lines(border.points[, 1], border.points[, 2], lwd = 2, type = "l", col = "black")
